#------------------------------------------------------------------------------
#
#   Interesting Times - Level
#
#------------------------------------------------------------------------------

import os
from random import choice
from numpy import ndarray
from albow.resource import get_sound
from tiles import NoTile
from team import Team
from probability import chance
from news_items import random_news_item
from vectors import vector
from sound import play_sound, bling
import constants as k

#------------------------------------------------------------------------------

class TileArray(object):

	def __init__(self, w, h):
		self.cells_wide = w
		self.cells_high = h
		self.size = vector(w, h)
		self.irange = xrange(w)
		self.jrange = xrange(h)
		self.cells = ndarray((w, h), 'O')
	
	def tile_at(self, coords):
		i = coords[0]
		j = coords[1]
		if 0 <= i <= self.cells_wide and 0 <= j < self.cells_high:
			return self.cells[i, j] or NoTile
		else:
			return NoTile
	
	def add_tile(self, tile, origin):
		tile.origin = origin
		self.remove_tile(tile)
		self.replace_cells(tile, tile)
	
	def remove_tile_at(self, cell):
		tile = self.tile_at(cell)
		if tile:
			self.remove_tile(tile)
	
	def remove_tile(self, tile):
		for cell in self.cells_covered_by_tile(tile):
			old_tile = self.tile_at(cell)
			self.cells[cell] = None
			if old_tile:
				self.remove_tile(old_tile)
	
	def replace_cells(self, tile, value):
		cells = self.cells
		for cell in self.cells_covered_by_tile(tile):
			cells[cell] = value
	
	def cells_covered_by_tile(self, tile):
		x0, y0 = tile.origin
		for x1 in tile.irange:
			x = x0 + x1
			if 0 <= x < self.cells_wide:
				for y1 in tile.jrange:
					y = y0 + y1
					if 0 <= y < self.cells_high:
						yield x, y
	
	def iter_tiles(self):
		cells = self.cells
		for i in self.irange:
			for j in self.jrange:
				tile = cells[i, j]
				if tile:
					p = tile.origin
					if i == p[0] and j == p[1]:
						yield tile
	
	def point_to_cell(self, p):
		return p // k.CELL_SIZE
	
	def cell_to_point(self, cell):
		return cell * k.CELL_SIZE
		
	def iter_adjacent_road_cells(self, cell):
		for d in k.all_directions:
			cell2 = cell + d
			if self.tile_at(cell2).is_road:
				yield cell2
	
	def iter_road_cells_around_tile(self, tile):
		for cell in tile.iter_cells(1, 1):
			if self.tile_at(cell).is_road:
				yield cell
	
	def random_cell(self):
		return random_vector(self.size)

#------------------------------------------------------------------------------

class City(TileArray):
	pass
		
#------------------------------------------------------------------------------

class Level(object):
	#  city           City
	#  accum_minutes  float
	#  ticks_per_minute float
	#  clock          int              game minutes since level start
	#  time           tuple            (days, hours, minutes)
	#  roads          [RoadTile]       all roads on level
	#  houses         [HouseTile]      all houses on level
	#  newsstands     [NewsStandTile]  all newsstands on level
	#  news_items     [NewsItem]
	#  completed      boolean
	#  runner_start_cells  [cell]   candidate runner start positions

	cells_wide = 50
	cells_high = 33
	
	breaking_news_sound = get_sound("telephone-ring.ogg")
	breaking_news_sound.set_volume(0.5)
	
	printing_sound = get_sound("printing-press.ogg")
	
	delivery_start_sound = get_sound("bicycle-bell.ogg")
	delivery_start_sound.set_volume(0.5)
	
	teams = None
	time = (1, 0, 0)
	selected_actor = None
	
	game_speed = property(
		lambda self: self.minutes_per_tick,
		lambda self, v: setattr(self, 'minutes_per_tick', v)
	)
	
	blue_team = property(lambda self: self.team(k.BLUE_TEAM))
	red_team = property(lambda self: self.team(k.RED_TEAM))
	
	def __init__(self, n, city):
		print "Level.__init__:", self, n, city ###
		self.city = City(self.cells_wide, self.cells_high)
		#self.runners = []
	
#	def __setstate__(self, d):
#		if 'runners' not in d:
#			d['runners'] = []
#		self.__dict__.update(d)
	
	def prepare(self):
		self.game_speed = k.SLOWEST_GAME_SPEED
		self.accum_minutes = 0.0
		h, m = k.START_TIME
		self.clock = h * 60 + m
		roads = []
		houses = []
		newsstands = []
		self.news_items = []
		for tile in self.city.iter_tiles():
			if tile.is_road:
				roads.append(tile)
			elif tile.is_house:
				houses.append(tile)
			elif tile.is_newsstand:
				newsstands.append(tile)
		self.roads = roads
		self.houses = houses
		self.newsstands = newsstands
		self.runner_start_cells = [
			cell for ns in self.newsstands
				for cell in self.city.iter_road_cells_around_tile(ns)]
		#print "Level: candidate runner start cells =", self.runner_start_cells ###
		for tile in self.houses:
			tile.prepare(self)
		for tile in self.newsstands:
			tile.prepare(self)
		self.teams = [Team(i + 1, self) for i in (0, 1)]
		if os.getenv("IT_AUTOPLAY"):
			self.blue_team.auto_play = True
		self.red_team.auto_play = True
		self.completed = False
		#p = (4, 4) ###
		#self.team(k.BLUE_TEAM).create_runner(p) ###
		#p = (10, 4) ###
		#self.team(k.BLUE_TEAM).create_reporter(p) ###

	def level_is_completed(self):
		return self.completed
	
	def team(self, s):
		return self.teams[s - 1]
		
	def level_is_completed(self):
		return False

	def begin_frame(self):
		#print "Level.begin_frame"
		self.update_clock()
		for team in self.teams:
			team.begin_frame()
	
	def update_clock(self):
		#t = self.tick_count + 1
		#self.tick_count = t
		m = self.accum_minutes + self.minutes_per_tick
		#if t == k.TICKS_PER_MINUTE:
		while m >= 1.0:
			#self.tick_count = 0
			m -= 1.0
			c = self.clock + 1
			self.clock = c
			c, minutes = divmod(c, 60)
			days, hours = divmod(c, 24)
			self.time = (days + 1, hours, minutes)
			self.generate_news()
			self.timed_events()
			self.check_for_win()
		self.accum_minutes = m
	
	def check_for_win(self):
		if self.all_houses_subscribed():
			self.completed = True
			bling("You have 100% readership! You own the town!")
	
	def all_houses_subscribed(self):
		s = k.BLUE_TEAM
		for h in self.houses:
			if h.subscription <> s:
				return False
		return True
	
	def generate_news(self):
		if chance(k.NEWS_GENERATION_RATE):
			#print "Level: generating news" ###
			roads = self.roads
			if roads:
				item = random_news_item(roads)
				self.news_items.append(item)
				play_sound(0, self.breaking_news_sound)
			else:
				print "*** Level: Can't generate news because there are no roads"
	
	def remove_news_item(self, item):
		self.news_items.remove(item)
	
	def timed_events(self):
		time_of_day = self.time[1:]
		#print "Level: time_of_day =", time_of_day ###
		if time_of_day == k.HOUSE_UPDATE_TIME:
			self.update_houses()
		if time_of_day == k.PRINTING_TIME:
			self.print_newspaper()
		if time_of_day == k.DELIVERY_START_TIME:
			self.start_delivery()
		if time_of_day == k.DELIVERY_END_TIME:
			self.stop_delivery()
		if time_of_day == k.PURCHASING_TIME:
			self.purchasing()
		if time_of_day == k.REPORTER_PAY_TIME:
			self.pay_reporters()
	
	def update_houses(self):
		for house in self.houses:
			house.update()
		for house in self.houses:
			house.decay_loyalties()
	
	def print_newspaper(self):
		print "Level: Time to print the paper!" ###
		self.red_team.print_newspaper()
		if self.blue_team.print_newspaper():
			self.printing_sound.play()
	
	def pay_reporters(self):
		for team in self.teams:
			team.pay_reporters()
	
	def start_delivery(self):
		print "Level.start_delivery" ###
		self.delivery_start_sound.play()
		for team in self.teams:
			team.pay_runners()
			team.start_delivery()
	
	def stop_delivery(self):
		print "Level.stop_delivery" ###
		for team in self.teams:
			team.stop_delivery()
	
	def purchasing(self):
		for house in self.houses:
			house.purchasing()

	def set_time_of_day(self, h, m):
		d = self.time[0] - 1
		self.clock = (d * 24 + h) * 60 + m
	
	def random_road_tile(self):
		roads = self.roads
		if roads:
			return choice(roads)

	def pick_runner_start_tile(self):
		cells = self.runner_start_cells
		if cells:
			return choice(cells)
